/*                
 *                        Licensing Notification
 *
 * Permission to use, reproduce, prepare derivative works, and to redistribute
 * to others this software and its documentation is hereby granted, provided
 * that this notice is retained thereon and on all copies or modifications. UChicago
 * Argonne, LLC, Tech-X Corporation, and all other contributors make no representations
 * as to the suitability and operability of this software for any purpose. It is
 * provided "as is" without express or implied warranty.
 *
 */

#ifndef TX_HYB_MATRIX_H
#define TX_HYB_MATRIX_H

#include <cusp/csr_matrix.h>
#include <cusp/hyb_matrix.h>
#include <tx_common.h>

#if defined(PETSC_USE_COMPLEX)
#if defined(PETSC_USE_REAL_SINGLE)  
#define cusparse_hyb_solve cusparseChybsv_solve
#define cusparse_hyb_analysis cusparseChybsv_analysis
#define cusparse_hyb_spmv cusparseChybmv
#define cusparse_hyb_convert cusparseCcsr2hyb
#elif defined(PETSC_USE_REAL_DOUBLE)
#define cusparse_hyb_solve cusparseZhybsv_solve
#define cusparse_hyb_analysis cusparseZhybsv_analysis
#define cusparse_hyb_spmv cusparseZhybmv
#define cusparse_hyb_convert cusparseZcsr2hyb
#endif
#else
#if defined(PETSC_USE_REAL_SINGLE)  
#define cusparse_hyb_solve cusparseShybsv_solve
#define cusparse_hyb_analysis cusparseShybsv_analysis
#define cusparse_hyb_spmv cusparseShybmv
#define cusparse_hyb_convert cusparseScsr2hyb
#elif defined(PETSC_USE_REAL_DOUBLE)
#define cusparse_hyb_solve cusparseDhybsv_solve
#define cusparse_hyb_analysis cusparseDhybsv_analysis
#define cusparse_hyb_spmv cusparseDhybmv
#define cusparse_hyb_convert cusparseDcsr2hyb
#endif
#endif

#define CUSPMATRIXHYB cusp::hyb_matrix<PetscInt,PetscScalar,cusp::device_memory>

// -----------------------------------------------------------------------------

/*
	Kernel function is based on CUSP library's spmv_hyb_kernel
	function which is licensed under the Apache 2.0 license.  Here
	is the license information for that part of the code.

 *  Copyright 2008-2012 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.

*/

// -----------------------------------------------------------------------------


/***********************/
/*   Hybrid Matrix     */
/***********************/
template <typename index_type, typename value_type>
class HybMatrix {
public:
  typedef CUSPMATRIXHYB MatrixType;
 
  HybMatrix() {
    cusparseA=0;
    cusparseATranspose=0;
    cusparseAIsAllocated=false;
    cusparseATransposeIsAllocated=false;
  }

  /* Destructor */
  virtual ~HybMatrix() {
    if(cusparseAIsAllocated) {
      cusparseStatus_t stat = cusparseDestroyHybMat(cusparseA);
      CUSPARSE_ERR_CHK_NORETURN(stat, "~HybMatrix()", "cusparseDestroyHybMat()")
    }      
    if(cusparseATransposeIsAllocated) {
      cusparseStatus_t stat = cusparseDestroyHybMat(cusparseATranspose);
      CUSPARSE_ERR_CHK_NORETURN(stat, "~HybMatrix()", "cusparseDestroyHybMat()")
    }      
  }

  // Get the matrix type;
  Tx_Mat_type_t getMatrixType() {
    return TX_HYB;
  }
 
  /* Set the Matrix explicitly on the device*/
  void setMatrix(const cusp::csr_matrix<index_type,value_type,cusp::host_memory>& matCPU) {
    A = matCPU;
  }  



  /* Set the Matrix */
  cusparseStatus_t setCusparseMatrix(cusparseHandle_t cusparseHandle, 
				     cusparseMatDescr_t cusparseMatDescription,
				     index_type numRows, index_type numCols, index_type nz,
				     index_type * row_offsets, index_type * column_indices,
				     value_type * values) {

    cusparseStatus_t stat=CUSPARSE_STATUS_SUCCESS;
    if (!cusparseA) {
      stat = cusparseCreateHybMat(&cusparseA);
      cusparseAIsAllocated=true;
      CUSPARSE_ERR_CHK(stat, "HybMatrix::setCusparseMatrix", "cusparseCreateHybMat()");
      // create a temporary 
      CUSPMATRIXCSR B; /* CSR version of the matrix */
      B.resize(numRows,numCols,nz);    
      B.row_offsets.assign(row_offsets, row_offsets+numRows+1);
      B.column_indices.assign(column_indices, column_indices+nz);
      B.values.assign(values, values+nz);

#if defined (PETSC_USE_64BIT_INDICES)
      CUSPINTARRAYGPU32 row_offsets_int = B.row_offsets;
      CUSPINTARRAYGPU32 column_indices_int = B.column_indices;
#endif
      stat = cusparse_ell_convert(cusparseHandle, numRows, numCols, 
				  cusparseMatDescription,
				  thrust::raw_pointer_cast(&B.values[0]),
#if defined (PETSC_USE_64BIT_INDICES)
				  thrust::raw_pointer_cast(&row_offsets_int[0]), 
				  thrust::raw_pointer_cast(&column_indices_int[0]),
#else
				  thrust::raw_pointer_cast(&B.row_offsets[0]),
				  thrust::raw_pointer_cast(&B.column_indices[0]),
#endif
				  cusparseA, 0, CUSPARSE_HYB_PARTITION_AUTO);
      CUSPARSE_ERR_CHK(stat, "HybMatrix::setCusparseMatrix", "cusparse_hyb_convert()");
    }
    return stat;
  }



  /* Set the Matrix ... for the symmetric case */
  cusparseStatus_t setCusparseSymmMatrix(cusparseHandle_t cusparseHandle, 
					 cusparseMatDescr_t cusparseMatDescription,
					 index_type numRows, index_type numCols, index_type nz,
					 index_type * row_offsets, index_type * column_indices,
					 value_type * values) {

    cusparseStatus_t stat=CUSPARSE_STATUS_SUCCESS;
    if (!cusparseA) {

      index_type * row_offsets_full;
      index_type * column_indices_full;
      value_type * values_full;
      row_offsets_full = (index_type *) malloc((numRows+1)*sizeof(index_type));
      memset(row_offsets_full, 0, (numRows+1)*sizeof(index_type));
      
      column_indices_full = (index_type *) malloc((2*nz-numRows)*sizeof(index_type));
      memset(column_indices_full, 0, (2*nz-numRows)*sizeof(index_type));
      
      values_full = (value_type *) malloc((2*nz-numRows)*sizeof(value_type));
      memset(values_full, 0, (2*nz-numRows)*sizeof(value_type));    
      
      /* Convert to a full matrix */
      convertSymmToFullMatrix<index_type, value_type>(numRows, nz,
						      row_offsets, column_indices, values,
						      row_offsets_full, column_indices_full, values_full);

      /* reset the type to general ... has to happen before csr to hyb conversion */
      stat = cusparseSetMatType(cusparseMatDescription, CUSPARSE_MATRIX_TYPE_GENERAL);
      CUSPARSE_ERR_CHK(stat, "setCusparseSymmMatrix()", "cusparseSetMatType()");

      /* build the full matrix */
      stat = setCusparseMatrix(cusparseHandle, cusparseMatDescription,
			       numRows, numCols, 2*nz-numRows,
			       row_offsets_full, column_indices_full, values_full);

      /* free the temp data */
      free(row_offsets_full);
      free(column_indices_full);
      free(values_full);
    }
    return stat;
  }


  /* Set the Matrix ... for the hermitian case */
  cusparseStatus_t setCusparseHermMatrix(cusparseHandle_t cusparseHandle, 
					 cusparseMatDescr_t cusparseMatDescription,
					 index_type numRows, index_type numCols, index_type nz,
					 index_type * row_offsets, index_type * column_indices,
					 value_type * values) {

    cusparseStatus_t stat=CUSPARSE_STATUS_MATRIX_TYPE_NOT_SUPPORTED;
    return stat;
  }


  /* Set the Matrix */
  cusparseStatus_t setCusparseMatrixTranspose(cusparseHandle_t cusparseHandle, 
					      cusparseMatDescr_t cusparseMatDescription,
					      index_type numRows, index_type numCols, index_type nz,
					      index_type * row_offsets, 
					      index_type * column_indices,
					      value_type * values) {

    cusparseStatus_t stat=CUSPARSE_STATUS_SUCCESS;
    if (!cusparseATranspose) {
      stat = cusparseCreateHybMat(&cusparseATranspose);
      cusparseATransposeIsAllocated=true;
      CUSPARSE_ERR_CHK(stat, "EllMatrix::setCusparseMatrixTranspose", "cusparseCreateHybMat()");
      // create a temporary 
      CUSPMATRIXCSR B; /* CSR version of the matrix */
      B.resize(numRows,numCols,nz);    
      B.row_offsets.assign(row_offsets, row_offsets+numRows+1);
      B.column_indices.assign(column_indices, column_indices+nz);
      B.values.assign(values, values+nz);
      
#if defined (PETSC_USE_64BIT_INDICES)
      CUSPINTARRAYGPU32 row_offsets_int = B.row_offsets;
      CUSPINTARRAYGPU32 column_indices_int = B.column_indices;
#endif
      
      // Build csc matrix on demand    
      CUSPARRAY cscValues(nz);
      CUSPINTARRAYGPU32 cscRowIndices(nz);
      CUSPINTARRAYGPU32 cscColumnOffsets(numCols+1);    
      stat=cusparse_csr2csc(cusparseHandle, numRows, numCols, nz,
			    thrust::raw_pointer_cast(B.values.data()),
#if defined (PETSC_USE_64BIT_INDICES)
			    thrust::raw_pointer_cast(&row_offsets_int[0]), 
			    thrust::raw_pointer_cast(&column_indices_int[0]),
#else
			    thrust::raw_pointer_cast(B.row_offsets.data()),
			    thrust::raw_pointer_cast(B.column_indices.data()),
#endif
			    thrust::raw_pointer_cast(cscValues.data()), 
			    thrust::raw_pointer_cast(cscRowIndices.data()), 
			    thrust::raw_pointer_cast(cscColumnOffsets.data()),
			    CUSPARSE_ACTION_NUMERIC, CUSPARSE_INDEX_BASE_ZERO);
      
      
      stat = cusparse_ell_convert(cusparseHandle, numRows, numCols, 
				  cusparseMatDescription,
				  thrust::raw_pointer_cast(cscValues.data()), 
				  thrust::raw_pointer_cast(cscColumnOffsets.data()),
				  thrust::raw_pointer_cast(cscRowIndices.data()), 
				  cusparseATranspose, 0, CUSPARSE_HYB_PARTITION_AUTO);
      CUSPARSE_ERR_CHK(stat, "EllMatrix::setCusparseMatrixTranpose", "cusparse_ell_convert()");
    }
    return stat;  
  }

  /* Set the Transpose of the Matrix ... for the symmetric case */
  cusparseStatus_t setCusparseSymmMatrixTranspose(cusparseHandle_t cusparseHandle, 
						  cusparseMatDescr_t cusparseMatDescription,
						  index_type numRows, index_type numCols, index_type nz,
						  index_type * row_offsets, 
						  index_type * column_indices,
						  value_type * values) {
    
    cusparseStatus_t stat=CUSPARSE_STATUS_SUCCESS;
    if (!cusparseATransposeIsAllocated) {
      index_type * row_offsets_full;
      index_type * column_indices_full;
      value_type * values_full;
      row_offsets_full = (index_type *) malloc((numRows+1)*sizeof(index_type));
      memset(row_offsets_full, 0, (numRows+1)*sizeof(index_type));
      
      column_indices_full = (index_type *) malloc((2*nz-numRows)*sizeof(index_type));
      memset(column_indices_full, 0, (2*nz-numRows)*sizeof(index_type));
      
      values_full = (value_type *) malloc((2*nz-numRows)*sizeof(value_type));
      memset(values_full, 0, (2*nz-numRows)*sizeof(value_type));    
      
      /* Convert to a full matrix */
      convertSymmToFullMatrix<index_type, value_type>(numRows, nz,
						      row_offsets, column_indices, values,
						      row_offsets_full, column_indices_full, values_full);
      
      /* reset the type to general ... has to happen before csr to hyb conversion */
      stat = cusparseSetMatType(cusparseMatDescription, CUSPARSE_MATRIX_TYPE_GENERAL);
      CUSPARSE_ERR_CHK(stat, "setCusparseSymmMatrixTranspose()", "cusparseSetMatType()");

      /* build the matrix */
      stat = setCusparseMatrixTranspose(cusparseHandle, cusparseMatDescription,
					numRows, numCols, 2*nz-numRows, row_offsets_full,
					column_indices_full, values_full);
            
      /* free the temp data */
      free(row_offsets_full);
      free(column_indices_full);
      free(values_full);
    }
    return stat;
  }

  /* Set the Transpose Matrix ... for the hermitian case */
  cusparseStatus_t setCusparseHermMatrixTranspose(cusparseHandle_t cusparseHandle, 
						  cusparseMatDescr_t cusparseMatDescription,
						  index_type numRows, index_type numCols, index_type nz,
						  index_type * row_offsets, 
						  index_type * column_indices,
						  value_type * values) {
    
    cusparseStatus_t stat=CUSPARSE_STATUS_MATRIX_TYPE_NOT_SUPPORTED;
    return stat;
  }



  /* Set the Matrix */
  cudaError_t setMatrix(index_type numRows, index_type numCols, index_type nz,
			index_type * row_offsets, index_type * column_indices,
			value_type * values) {

    // create a temporary 
    CUSPMATRIXCSR B; /* CSR version of the matrix */
    B.resize(numRows,numCols,nz);    
    B.row_offsets.assign(row_offsets, row_offsets+numRows+1);
    B.column_indices.assign(column_indices, column_indices+nz);
    B.values.assign(values, values+nz);
    // Let CUSP do the conversion
    A = B;
    return cudaGetLastError();
  }

  /* Get the Matrix */
  virtual int getCsrMatrix(CUSPMATRIXCSR** A) {
    CUSPARSE_ERR_CHK(CUSPARSE_STATUS_MATRIX_TYPE_NOT_SUPPORTED, "GPU_Matrix::getCsrMatrix()", "not supported for HYB matrix type");
  }

  /* Set the Indices For cprow==true (compressed row) */
  void setCPRowIndices(const index_type * indices, const index_type m) {
    // does nothing as of now
  }

  /* Set the inode data for when inode.use==true (inodes) */
  void setInodeData(const index_type * inodesCPU, const int inodeCountPlusOne, const int nnzPerRowMaxCPU,
                            const int nodeMaxCPU, const int nodeCountCPU) {
    // does nothing as of now
  }

  /* cusparse SpMV */
  int cusparseMultiply(const cusparseHandle_t cusparseHandle, 
		       const cusparseMatDescr_t cusparseMatDescription,
		       const CUSPARRAY *x, CUSPARRAY *y) {

    cusparseStatus_t stat=CUSPARSE_STATUS_SUCCESS;
#if defined(PETSC_USE_COMPLEX)
#if defined(PETSC_USE_REAL_DOUBLE)
    cuDoubleComplex alpha = {1.0, 0.0};
    cuDoubleComplex beta = {0.0, 0.0};
#else
    cuFloatComplex alpha = {1.0f, 0.0f};
    cuFloatComplex beta = {0.0f, 0.0f};
#endif
#else
#if defined(PETSC_USE_REAL_DOUBLE)
    PetscScalar alpha = 1.0;
    PetscScalar beta = 0.0;
#else
    PetscScalar alpha = 1.0f;
    PetscScalar beta = 0.0f;
#endif
#endif
    if (cusparseA) 
      stat = cusparse_hyb_spmv(cusparseHandle, CUSPARSE_OPERATION_NON_TRANSPOSE,
			       &alpha, cusparseMatDescription, cusparseA,
			       thrust::raw_pointer_cast(x->data()), &beta,			    
			       thrust::raw_pointer_cast(y->data()));    
    CUSPARSE_ERR_CHK(stat, "HybMatrix::cusparseMultiply", "cusparse_hyb_spmv()");
    return (int) stat;
  }


  /* cusparse SpMV */
  int cusparseMultiplyTranspose(const cusparseHandle_t cusparseHandle, 
				const cusparseMatDescr_t cusparseMatDescription,
				const CUSPARRAY *x, CUSPARRAY *y) {

    cusparseStatus_t stat=CUSPARSE_STATUS_SUCCESS;
#if defined(PETSC_USE_COMPLEX)
#if defined(PETSC_USE_REAL_DOUBLE)
  cuDoubleComplex alpha = {1.0, 0.0};
  cuDoubleComplex beta = {0.0, 0.0};
#else
  cuFloatComplex alpha = {1.0f, 0.0f};
  cuFloatComplex beta = {0.0f, 0.0f};
#endif
#else
#if defined(PETSC_USE_REAL_DOUBLE)
  PetscScalar alpha = 1.0;
  PetscScalar beta = 0.0;
#else
  PetscScalar alpha = 1.0f;
  PetscScalar beta = 0.0f;
#endif
#endif
  if (cusparseATranspose)
    stat = cusparse_ell_spmv(cusparseHandle, CUSPARSE_OPERATION_NON_TRANSPOSE,
			     &alpha, cusparseMatDescription, cusparseATranspose,
			     thrust::raw_pointer_cast(x->data()), &beta,			    
			     thrust::raw_pointer_cast(y->data()));    
    CUSPARSE_ERR_CHK(stat, "HybMatrix::cusparseMultiplyTranspose", "cusparse_hyb_spmv()");
    return (int) stat;
  }


  /* SpMV with our CUSP-like kernels */
  int multiply(const cudaStream_t stream, const CUSPARRAY *x, CUSPARRAY *y) {
    return 0;
  }

  /* Solve Analysis */
  cusparseStatus_t solveAnalysis(cusparseHandle_t cusparseHandle,
				 cusparseMatDescr_t cusparseMatDescription,
				 cusparseSolveAnalysisInfo_t cusparseSolveAnalysisInfo,
				 cusparseOperation_t OPERATION) {
    
    cusparseStatus_t stat = cusparse_hyb_analysis(cusparseHandle, OPERATION,
						  cusparseMatDescription, cusparseA,
						  cusparseSolveAnalysisInfo);
    CUSPARSE_ERR_CHK(stat, "HybMatrix::solveAnalysis", "cusparse_hyb_analysis()");
    return stat;
  };
  

  /* solve */
  cusparseStatus_t solve(cusparseHandle_t cusparseHandle,
			 cusparseMatDescr_t cusparseMatDescription,
			 cusparseSolveAnalysisInfo_t cusparseSolveAnalysisInfo,
			 cusparseOperation_t OPERATION,
			 const CUSPARRAY *bGPU, CUSPARRAY *xGPU) {

#if defined(PETSC_USE_COMPLEX)
#if defined(PETSC_USE_REAL_DOUBLE)
  cuDoubleComplex alpha = {1.0, 0.0};
#else
  cuFloatComplex alpha = {1.0f, 0.0f};
#endif
#else
#if defined(PETSC_USE_REAL_DOUBLE)
  PetscScalar alpha = 1.0;
#else
  PetscScalar alpha = 1.0f;
#endif
#endif
    cusparseStatus_t stat = cusparse_hyb_solve(cusparseHandle, OPERATION,
					       &alpha, cusparseMatDescription, cusparseA,
					       cusparseSolveAnalysisInfo,
					       (const PetscScalar *) thrust::raw_pointer_cast(bGPU->data()),
					       (PetscScalar *) thrust::raw_pointer_cast(xGPU->data()));
    CUSPARSE_ERR_CHK(stat, "HybMatrix::solve()", "cusparse_hyb_solve()");
    return stat;
  }

  /* Solve Analysis with transpose of the matrix */
  cusparseStatus_t solveAnalysisTranspose(cusparseHandle_t cusparseHandle,
					  cusparseMatDescr_t cusparseMatDescription,
					  cusparseSolveAnalysisInfo_t cusparseSolveAnalysisInfo,
					  cusparseOperation_t OPERATION) {
    CUSPARSE_ERR_CHK(CUSPARSE_STATUS_MATRIX_TYPE_NOT_SUPPORTED, "HybMatrix::solveAnalysisTranspose()", "hyb matrix type not supported : ")
  };
  

  /* solve with transpose of the matrix  */
  cusparseStatus_t solveTranspose(cusparseHandle_t cusparseHandle,
				  cusparseMatDescr_t cusparseMatDescription,
				  cusparseSolveAnalysisInfo_t cusparseSolveAnalysisInfo,
				  cusparseOperation_t OPERATION,
				  const CUSPARRAY *bGPU, CUSPARRAY *xGPU) {
    CUSPARSE_ERR_CHK(CUSPARSE_STATUS_MATRIX_TYPE_NOT_SUPPORTED, "HybMatrix::solveTranspose()", "hyb matrix type not supported : ")
  }


private:
  MatrixType A;/* container for the matrix */
  cusparseHybMat_t cusparseA; /* cusparse container for the matrix */
  cusparseHybMat_t cusparseATranspose; /* cusparse container for the matrix transpose */
  bool cusparseAIsAllocated;
  bool cusparseATransposeIsAllocated;
};

#endif // TX_HYB_MATRIX_H
